from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import Storage
from django.utils.module_loading import import_string

from .question_types import QuestionTypeRegistry


upload_storage = None
upload_to_handler = None
upload_allowed_file_extensions = tuple()
active_question_types = list()

# validate & load QUESTIONNAIRE_CORE_UPLOAD_STORAGE
if getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_STORAGE', None):
    try:
        storage_class = import_string(getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_STORAGE'))
        upload_storage = storage_class()
        if not isinstance(upload_storage, Storage):
            raise ImproperlyConfigured(
                'User-defined QUESTIONNAIRE_CORE_UPLOAD_STORAGE is not a subclass of django.core.files.storage.Storage'
            )
    except ImportError:
        raise ImproperlyConfigured('Could not import user-defined QUESTIONNAIRE_CORE_UPLOAD_STORAGE')

# validate & load QUESTIONNAIRE_CORE_UPLOAD_TO_HANDLER
if getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_TO_HANDLER', None):
    try:
        upload_to_handler = import_string(getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_TO_HANDLER'))
        if not callable(upload_to_handler):
            raise ImproperlyConfigured('User-defined QUESTIONNAIRE_CORE_UPLOAD_TO_HANDLER is not callable.')
    except ImportError:
        raise ImproperlyConfigured('Could not import user-defined QUESTIONNAIRE_CORE_UPLOAD_TO_HANDLER')

# validate QUESTIONNAIRE_CORE_UPLOAD_ALLOWED_FILE_EXTENSIONS
if getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_ALLOWED_FILE_EXTENSIONS', None):
    if not isinstance(getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_ALLOWED_FILE_EXTENSIONS', list()), (list, tuple)):
        raise ImproperlyConfigured(
            'User-defined QUESTIONNAIRE_CORE_UPLOAD_ALLOWED_FILE_EXTENSIONS must be a list or tuple.'
        )
    else:
        upload_allowed_file_extensions = getattr(settings, 'QUESTIONNAIRE_CORE_UPLOAD_ALLOWED_FILE_EXTENSIONS')

# validate & load QUESTIONNAIRE_CORE_ENABLED_QUESTION_TYPES
if getattr(settings, 'QUESTIONNAIRE_CORE_ENABLED_QUESTION_TYPES', None):
    enable_types = getattr(settings, 'QUESTIONNAIRE_CORE_ENABLED_QUESTION_TYPES')
    for enable_type in enable_types:
        if '.' in enable_type:
            try:
                active_type = import_string(enable_type)
                active_question_types.append(active_type)
            except ImportError:
                raise ImproperlyConfigured('Could not import user-defined question type: {}'.format(enable_type))
        else:
            active_type = QuestionTypeRegistry.get_question_type(enable_type)
            if active_type:
                active_question_types.append(active_type)
            else:
                raise ImproperlyConfigured('Invalid name for builtin question type: {}'.format(enable_type))
else:
    # default: enable all builtin question types
    active_question_types = QuestionTypeRegistry.get_question_types().values()



from django.core.validators import FileExtensionValidator
from django.db import models

from . import settings


class QuestionTypeField(models.CharField):
    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # don't serialize kwargs set at runtime
        kwargs.pop('choices', None)
        return name, path, args, kwargs


class DynamicStorageFileField(models.FileField):
    def __init__(self, verbose_name=None, name=None, upload_to='', storage=None, **kwargs):
        if not storage and settings.upload_storage:
            storage = settings.upload_storage
        if not upload_to and settings.upload_to_handler:
            upload_to = settings.upload_to_handler
        if settings.upload_allowed_file_extensions:
            if not kwargs.get('validators'):
                kwargs['validators'] = list()
            kwargs['validators'].append(
                FileExtensionValidator(allowed_extensions=settings.upload_allowed_file_extensions)
            )
        super().__init__(verbose_name, name, upload_to, storage, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # don't serialize kwargs set at runtime by project settings
        kwargs.pop('storage', None)
        kwargs.pop('upload_to', None)
        kwargs.pop('validators', None)
        return name, path, args, kwargs



import json

from django.contrib import admin
from django.contrib.postgres import fields as postgres_fields
from django.forms import widgets

from ordered_model.admin import OrderedTabularInline

from .models import Question, QuestionAnswer, Questionnaire, QuestionnaireResult


try:
    from ordered_model.admin import OrderedInlineModelAdminMixin  # v3+
except ImportError:
    class OrderedInlineModelAdminMixin(object):
        def get_urls(self):
            urls = super().get_urls()
            for inline in self.inlines:
                if hasattr(inline, 'get_urls'):
                    urls = inline.get_urls(self) + urls
            return urls


class PrettyJSONWidget(widgets.Textarea):
    def format_value(self, value):
        try:
            return json.dumps(json.loads(value), indent=2)  # reformat json
        except TypeError:
            return value


class QuestionnaireQuestionListModelInline(OrderedTabularInline):
    model = Question
    fields = ('question_type', 'question_text', 'question_options', 'required', 'order', 'move_up_down_links',)
    readonly_fields = ('order', 'move_up_down_links',)
    extra = 1
    ordering = ('order',)
    formfield_overrides = {
        postgres_fields.JSONField: {
            'widget': PrettyJSONWidget
        },
    }


class QuestionnaireAnswerListModelInline(OrderedTabularInline):
    model = QuestionAnswer
    fields = ('question', 'answer_data', 'order', 'move_up_down_links',)
    readonly_fields = ('order', 'move_up_down_links',)
    extra = 0
    ordering = ('order',)


@admin.register(Questionnaire)
class QuestionnaireAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    list_display = ('title', )
    inlines = (QuestionnaireQuestionListModelInline,)


@admin.register(Question)
class QuestionAdmin(admin.ModelAdmin):
    pass


@admin.register(QuestionnaireResult)
class QuestionnaireResultAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    readonly_fields = ('created_at', 'updated_at')
    inlines = (QuestionnaireAnswerListModelInline,)


@admin.register(QuestionAnswer)
class QuestionAnswerAdmin(admin.ModelAdmin):
    pass



import inspect
import string

from django import forms
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.template.engine import Engine
from django.template.exceptions import TemplateDoesNotExist


class QuestionTypeRegistry:
    """Registry for question type classes."""
    _registered_types = {}

    @classmethod
    def register(cls, question_type):
        name = question_type.meta.name
        cls._registered_types[name] = question_type

    @classmethod
    def unregister(cls, question_type):
        name = question_type.meta.name
        if name in cls._registered_types:
            del cls._registered_types[name]

    @classmethod
    def get_question_types(cls):
        return cls._registered_types

    @classmethod
    def get_question_type(cls, name):
        try:
            return cls._registered_types[name]
        except KeyError:
            return


class Options:
    """Class to hold `meta` options of a question type class."""
    REQUIRED = ('name', 'verbose_name')
    NAME_VALID_CHARS = string.ascii_lowercase + string.digits + '_'

    def __init__(self, cls, meta):
        for required_option in self.REQUIRED:
            if not meta.get(required_option, None):
                raise AttributeError('{}.Meta missing required field "{}"'.format(cls.__name__, required_option))

        if not set(self.NAME_VALID_CHARS).issuperset(set(meta.get('name'))):
            raise ValueError(
                'Invalid name for question type class {}: {} (valid chars: {})'.format(
                    cls.__name__,
                    meta.get('name'),
                    self.NAME_VALID_CHARS,
                )
            )

        self.name = meta.get('name')
        self.verbose_name = meta.get('verbose_name')
        self.multiple = meta.get('multiple', False)
        self.widget_class = meta.get('widget_class', None)
        self._widget_template_name = meta.get('widget_template_name')
        self._widget_option_template_name = meta.get('widget_option_template_name')

    @property
    def widget_template_name(self):
        return self._widget_template_name or self.select_default_template('template_name')

    @property
    def widget_option_template_name(self):
        return self._widget_option_template_name or self.select_default_template('option_template_name')

    def select_default_template(self, template_key):
        """Select default template for the formfield widget.

        Returns the packaged template for the widget if available.
        """
        if template_key == 'template_name':
            default_template = 'questionnaire_core/widgets/{name}.html'.format(name=self.name)
        elif template_key == 'option_template_name':
            default_template = 'questionnaire_core/widgets/{name}_option.html'.format(name=self.name)
        else:
            return

        try:
            template_engine = Engine.get_default()
            packaged_template = template_engine.get_template(default_template)
            return packaged_template.name
        except TemplateDoesNotExist:
            return


class QuestionTypeMeta(type):
    """Meta class for question type classes.

    Responsible for registering question type classes with QuestionTypeRegistry
    and setting up the `meta` attribute of question type classes.
    """
    def __new__(mcs, name, bases, attrs):
        super_new = super(QuestionTypeMeta, mcs).__new__

        # register only subclasses of QuestionTypeBase not QuestionTypeBase itself
        if name == 'QuestionTypeBase':
            return super_new(mcs, name, bases, attrs)

        attr_meta = attrs.pop('Meta', None)

        # don't register abstract classes
        if getattr(attr_meta, 'abstract', False):
            return super_new(mcs, name, bases, attrs)

        if not attr_meta or not inspect.isclass(attr_meta):
            raise AttributeError('{}.Meta attribute missing or not a class'.format(name))

        new_class = super_new(mcs, name, bases, attrs)

        # create meta attribute (instance of Options) from new_class.Meta (similar to Model._meta)
        setattr(new_class, 'meta', Options(new_class, attr_meta.__dict__))

        QuestionTypeRegistry.register(new_class)

        return new_class


class QuestionTypeBase(object, metaclass=QuestionTypeMeta):
    """Base class for question type classes"""

    class OptionsForm(forms.Form):
        pass

    def __init__(self, question):
        self.question = question

    def question_option_form(self, *args, **kwargs):  # arg0: request (optional)
        return self.OptionsForm

    @classmethod
    def question_option_form_fields(cls):
        return cls.OptionsForm.base_fields

    def clean_question_options(self, question_options):
        """Clean question options (`Question.question_options`).

        Override to implement any custom validations of the question options of the question type.

        :param question_options: django admin form field data
        :type question_options: dict
        :return: cleaned form field data
        :rtype: dict
        :raises: django.forms.ValidationError: Validation error
        """
        return question_options

    def clean_answer_data(self, data):
        """Clean answer data (`QuestionAnswer.answer_data`).

        :param data: data returned from the formfield
        :return: cleaned formfield data
        :raises: django.forms.ValidationError: Validation error
        """
        return data

    def formfield(self, result_set):
        """Form field for the question type.

        :param result_set: result set of the current form
        :type result_set: questionnaire_core.models.QuestionnaireResult
        :return: django form field for the question type
        :rtype: django.forms.Field
        """
        raise NotImplementedError

    def formfield_widget(self, **kwargs):
        """Setup and return the widget for the formfield."""
        widget_attrs = self.formfield_widget_attrs()
        if 'attrs' in kwargs:
            widget_attrs.update(kwargs.get('attrs'))
        kwargs['attrs'] = widget_attrs
        widget = self.widget_class()(**kwargs)
        # set template attribute(s) of the widget
        for template_key in ('template_name', 'option_template_name'):
            meta_template_key = 'widget_{}'.format(template_key)
            if getattr(self.question.question_type_obj.meta, meta_template_key) and hasattr(widget, template_key):
                setattr(widget, template_key, getattr(self.question.question_type_obj.meta, meta_template_key))
        return widget

    def widget_class(self):
        """Return the configured widget class for the formfield."""
        if not self.meta.widget_class:
            raise ValueError('{}.Meta.widget_class attribute is missing.'.format(self.__class__.__name__))
        return self.meta.widget_class

    def formfield_widget_attrs(self):
        """Setup and return the attributes for the formfield widget (based on question options)."""
        attrs = dict()
        if 'autocomplete' in self.question.question_options:
            attrs.update({'autocomplete': 'on' if self.question.question_options.get('autocomplete') else 'off'})
        return attrs

    def initial_field_value(self, result_set):
        """Return the initial formfield value based on the supplied result set."""
        if self.question.question_type_obj.meta.multiple:
            return list(result_set.answers.filter(question=self.question).values_list('answer_data', flat=True))
        else:
            try:
                answer = result_set.answers.get(question=self.question)
                return answer.answer_data
            except ObjectDoesNotExist:
                return None
            except MultipleObjectsReturned:
                raise ValueError('Multiple answers found for QuestionType with multiple=False')

    def save(self, result_set, answer_data):
        from ..models import QuestionAnswer

        if self.meta.multiple:
            assert isinstance(answer_data, list)
            for answer_data_part in answer_data:
                QuestionAnswer.objects.create(
                    result_set=result_set,
                    question=self.question,
                    answer_data=answer_data_part,
                )
        else:
            QuestionAnswer.objects.create(
                result_set=result_set,
                question=self.question,
                answer_data=answer_data,
            )



import os

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.forms import widgets

from .base import QuestionTypeBase


class CustomFileInput(widgets.ClearableFileInput):
    input_type = 'file'

    def format_value(self, value):
        """
        Return the file object if it has a defined url attribute.
        """
        if self.is_initial(value):
            value.filename = os.path.basename(value.file.name)  # add filename attribute for usage in widget
            return value


class FileUpload(QuestionTypeBase):
    class Meta:
        name = 'file_upload'
        verbose_name = 'File upload'
        widget_class = CustomFileInput

    def formfield(self, result_set):
        required = self.question.required

        # disable required flag if field was originally required and file has already been uploaded
        if self.question.required:
            try:
                result_set.answers.get(question=self.question)
            except ObjectDoesNotExist:
                pass
            else:
                required = False

        return forms.FileField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=required,
        )

    def initial_field_value(self, result_set):
        try:
            answer = result_set.answers.get(question=self.question)
            return answer.file_upload.file
        except ObjectDoesNotExist:
            return None

    def save(self, result_set, file_upload):
        from ..models import AnswerFile, QuestionAnswer

        if file_upload is None:
            return

        try:
            answer = result_set.answers.get(question=self.question)
        except ObjectDoesNotExist:
            answer = None

        with transaction.atomic():
            # remove uploaded file ("clear" checkbox checked)
            if file_upload is False and answer is not None:
                answer.file_upload.delete()
                answer.delete()
                return

            # save some metadata in answer_data
            file_upload_name = os.path.basename(file_upload.name)
            answer_data = {
                'name': file_upload_name,
                'size': file_upload.size,
            }
            if not answer:
                answer = QuestionAnswer.objects.create(
                    result_set=result_set,
                    question=self.question,
                    answer_data=answer_data,
                )
            else:
                answer.answer_data = answer_data
                answer.save()

            if not hasattr(answer, 'file_upload'):
                AnswerFile.objects.create(answer=answer, file=file_upload)
            else:
                # delete previously uploaded file from storage & update file_upload field
                answer.file_upload.file.delete(save=False)
                answer.file_upload.file.save(file_upload_name, file_upload, save=True)



from .base import QuestionTypeRegistry
from .boolean import Boolean, BooleanYesNo
from .choices import Choices, ChoicesMultiple
from .file_upload import FileUpload
from .multiple import MultipleText
from .number import NumberDecimal, NumberInteger, NumberPercent
from .range import RangeSlider
from .text import TextLong, TextShort


__all__ = (
    'QuestionTypeRegistry',
    'Boolean',
    'BooleanYesNo',
    'Choices',
    'ChoicesMultiple',
    'FileUpload',
    'MultipleText',
    'NumberDecimal',
    'NumberInteger',
    'NumberPercent',
    'RangeSlider',
    'TextShort',
    'TextLong',
)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('questionnaire_core', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='answerfile',
            name='order',
            field=models.PositiveIntegerField(db_index=True, editable=False, verbose_name='order'),
        ),
        migrations.AlterField(
            model_name='question',
            name='order',
            field=models.PositiveIntegerField(db_index=True, editable=False, verbose_name='order'),
        ),
        migrations.AlterField(
            model_name='questionanswer',
            name='order',
            field=models.PositiveIntegerField(db_index=True, editable=False, verbose_name='order'),
        ),
    ]



from django import forms

from .base import QuestionTypeBase


class TextBase(QuestionTypeBase):
    class Meta:
        abstract = True

    class OptionsForm(forms.Form):
        min_length = forms.IntegerField(required=False)
        max_length = forms.IntegerField(required=False)

    def clean_question_options(self, question_options):
        """
        expected question_options format:
        {
            min_length: 10,
            max_length: 100
        }
        """

        if 'min_length' in question_options:
            try:
                question_options['min_length'] = int(question_options['min_length'])
            except ValueError:
                raise forms.ValidationError('value for "min_length" is not an integer')

        if 'max_length' in question_options:
            try:
                question_options['max_length'] = int(question_options['max_length'])
            except ValueError:
                raise forms.ValidationError('value for "max_length" is not an integer')

        return question_options


class TextShort(TextBase):
    class Meta:
        name = 'text_short'
        verbose_name = 'Text (Short)'
        widget_class = forms.TextInput

    def formfield(self, result_set):
        min_length = self.question.question_options.get('min_length')
        max_length = self.question.question_options.get('max_length')

        return forms.CharField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            min_length=min_length,
            max_length=max_length,
        )


class TextLong(TextBase):
    class Meta:
        name = 'text_long'
        verbose_name = 'Text (Long)'
        widget_class = forms.Textarea

    def formfield(self, result_set):
        min_length = self.question.question_options.get('min_length')
        max_length = self.question.question_options.get('max_length')

        return forms.CharField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            min_length=min_length,
            max_length=max_length,
        )



from django.views.generic import FormView


class QuestionnaireFormView(FormView):
    """Generic view for questionnaires."""
    form_class = None

    def get_questionnaire(self):
        """Return Questionnaire model for current form.

        The form is automatically generated based on the model instance.

        :return: related Questionnaire model
        :rtype: questionnaire_core.models.Questionnaire
        """
        raise NotImplementedError

    def get_questionnaire_result_set(self):
        """Return QuestionnaireResult model for current form.

        The results of the questionnaire are saved in the result set.

        Return an existing result set to allow editing
        or return a new (unsaved) model instance of QuestionnaireResult.

        :return: related QuestionnaireResult model
        :rtype: questionnaire_core.models.QuestionnaireResult
        """
        raise NotImplementedError

    def get_initial(self):
        return self.get_questionnaire_result_set().initial_form_data()

    def get_form_class(self):
        return self.get_questionnaire().build_form_class(self.get_questionnaire_result_set())

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)



from collections import OrderedDict

from django.contrib.postgres.fields import JSONField
from django.db import models, transaction
from django.utils.text import Truncator

from ordered_model.models import OrderedModel

from ..fields import QuestionTypeField
from ..forms import QuestionnaireFormBase
from ..question_types import QuestionTypeRegistry


def builtin_question_types():
    for question_type in sorted(QuestionTypeRegistry.get_question_types().values(), key=lambda q: q.meta.verbose_name):
        yield (question_type.meta.name, question_type.meta.verbose_name)


class QuestionnaireManager(models.Manager):
    def with_questions(self):
        return self.get_queryset().prefetch_related('questions')


class Questionnaire(models.Model):
    title = models.CharField(max_length=200)

    objects = QuestionnaireManager()

    def build_form_class(self, result_set):
        """Create form class for the questionnaire.

        :param result_set: result set used to save results and build initial form values
        :type result_set: questionnaire_core.models.QuestionnaireResult
        :return: form class for the questionnaire
        :rtype: QuestionnaireForm
        """
        question_form_dict = OrderedDict()

        # add question fields
        for question in self.questions.all():
            question_form_id = 'q{}'.format(question.pk)
            question_form_dict[question_form_id] = question.question_type_obj.formfield(result_set)

        # bind questionnaire & result_set to form
        question_form_dict['current_questionnaire'] = self
        question_form_dict['current_result_set'] = result_set

        form = type('QuestionnaireForm', (QuestionnaireFormBase,), question_form_dict)

        return form

    def __str__(self):
        return self.title

    def copy(self, title):
        questions = list(self.questions.all())

        with transaction.atomic():
            questionnaire_copy = Questionnaire.objects.create(
                title=title,
            )
            for question in questions:
                question.pk = None
                question.questionnaire = questionnaire_copy
                question.save()

        return questionnaire_copy


class Question(OrderedModel):
    questionnaire = models.ForeignKey(Questionnaire, related_name='questions', on_delete=models.CASCADE)
    question_text = models.CharField(max_length=5000)
    question_type = QuestionTypeField(max_length=32, choices=builtin_question_types())
    question_options = JSONField(blank=True, default=dict)
    required = models.BooleanField(default=True)

    order_with_respect_to = 'questionnaire'

    class Meta(OrderedModel.Meta):
        pass

    def question_type_class(self):
        if self.question_type:
            return QuestionTypeRegistry.get_question_type(self.question_type)
        raise ValueError('question_type not set')

    @property
    def question_type_obj(self):
        return self.question_type_class()(question=self)

    def __str__(self):
        return Truncator(self.question_text).chars(20)



from django import forms
from django.db import transaction


class QuestionnaireFormBase(forms.Form):
    """Base form for questionnaires.

     Fields are added dynamically by `Questionnaire.build_form_class()`.
    """

    def __init__(self, *args, **kwargs):
        # Property which can optionally be passed into form which sets all fields to disabled.
        readonly_form = kwargs.pop('is_readonly', False)

        super().__init__(*args, **kwargs)

        if readonly_form:
            for field_id, field in self.fields.items():
                field.disabled = True

        if self.data:
            for field_id, field in self.fields.items():
                if hasattr(field, 'dynamic_count'):
                    field_num = field.get_multi_field_count(self.data, field_id)
                    if field_num > 0:
                        self.fields[field_id] = field._question_type.formfield(
                            self.current_result_set,
                            min_fields=field_num,
                        )

    def clean(self):
        # per field (question/answer) validation
        for question in self.current_questionnaire.questions.all():
            field_id = 'q{}'.format(question.pk)
            answer_data = self.cleaned_data.get(field_id)
            try:
                self.cleaned_data[field_id] = question.question_type_obj.clean_answer_data(answer_data)
            except forms.ValidationError as e:
                self.add_error(field_id, e)

        return super().clean()

    @transaction.atomic
    def save(self, result_meta=None):
        result_set = self.current_result_set
        if result_set.pk:
            # on update remove previous answers (except uploaded files)
            result_set.answers.exclude(question__question_type__istartswith='file').delete()
        else:
            if result_meta is not None:
                result_set.result_meta = result_meta
            result_set.save()

        for question in self.current_questionnaire.questions.all():
            answer_data = self.cleaned_data.get('q{}'.format(question.pk))
            question.question_type_obj.save(result_set, answer_data)

        return result_set



from django import forms
from django.utils.translation import ugettext_lazy as _

from .base import QuestionTypeBase


class Boolean(QuestionTypeBase):
    class Meta:
        name = 'boolean'
        verbose_name = 'Boolean (Checkbox)'
        widget_class = forms.CheckboxInput

    def formfield(self, result_set):
        return forms.BooleanField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
        )


class BooleanYesNo(QuestionTypeBase):
    class Meta:
        name = 'boolean_yesno'
        verbose_name = 'Boolean (Yes/No)'
        widget_class = forms.RadioSelect

    def formfield(self, result_set):
        return forms.TypedChoiceField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            coerce=lambda x: x == 'True',
            choices=((True, _('Yes')), (False, _('No'))),
        )



from django import forms
from django.forms import widgets

from .base import QuestionTypeBase


class SliderInput(widgets.Input):
    input_type = 'range'


class RangeSlider(QuestionTypeBase):
    class Meta:
        name = 'range_slider'
        verbose_name = 'Range (slider)'
        widget_class = SliderInput

    class OptionsForm(forms.Form):
        min = forms.IntegerField(required=True)
        max = forms.IntegerField(required=True)
        step = forms.IntegerField(required=True)

    def clean_question_options(self, question_options):
        """
        expected question_options format:
        {
            min: 1,
            max: 10,
            step: 1
        }
        """

        if 'min' not in question_options:
            raise forms.ValidationError('key "min" required')
        try:
            question_options['min'] = int(question_options['min'])
        except ValueError:
            raise forms.ValidationError('value for "min" is not an integer')

        if 'max' not in question_options:
            raise forms.ValidationError('key "max" required')
        try:
            question_options['max'] = int(question_options['max'])
        except ValueError:
            raise forms.ValidationError('value for "max" is not an integer')

        if question_options['min'] >= question_options['max']:
            raise forms.ValidationError('value for "min" greater or equal to "max"')

        if 'step' not in question_options:
            raise forms.ValidationError('key "step" required')
        if question_options.get('step'):
            try:
                question_options['step'] = int(question_options['step'])
            except ValueError:
                raise forms.ValidationError('value for "step" is not an integer')

        if question_options['step'] > question_options['max']:
            raise forms.ValidationError('value for "step" greater than "max"')

        if question_options.get('initial'):
            try:
                question_options['initial'] = int(question_options['initial'])
            except ValueError:
                raise forms.ValidationError('value for "initial" is not an integer')

        return question_options

    def initial_field_value(self, result_set):
        initial = super().initial_field_value(result_set)
        return initial or self.question.question_options.get('initial')

    def formfield_widget_attrs(self):
        attrs = {
            'min': self.question.question_options.get('min'),
            'max': self.question.question_options.get('max'),
            'step': self.question.question_options.get('step', 1),
        }
        attrs.update(super().formfield_widget_attrs())
        return attrs

    def formfield(self, result_set):
        min_value = self.question.question_options.get('min')
        max_value = self.question.question_options.get('max')

        return forms.IntegerField(
            widget=self.formfield_widget(),
            min_value=min_value,
            max_value=max_value,
            required=self.question.required,
            label=self.question.question_text,
        )



from django import forms
from django.forms import widgets

from .base import QuestionTypeBase


class MultipleTextWidget(widgets.MultiWidget):
    def decompress(self, value):
        return value


class MultipleTextField(forms.MultiValueField):
    # custom property indicating field needs to be recreated depending on POST data
    dynamic_count = True

    def __init__(self, fields, *args, **kwargs):
        self._question_type = kwargs.pop('question_type', None)
        super().__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        # drop empty entries
        return [v for v in data_list if v]

    def clean(self, value):
        compressed_value = super().clean(value)
        question = self._question_type.question
        if len(compressed_value) < question.question_options.get('min', 1 if question.required else 0):
            raise forms.ValidationError(self.error_messages['required'], code='required')
        return compressed_value

    def get_multi_field_count(self, post_data, field_id):
        count = 0
        subfield_prefix = '{}_'.format(field_id)
        for subfield_id, field in post_data.items():
            if subfield_id.startswith(subfield_prefix):
                count += 1
        return count


class MultipleText(QuestionTypeBase):
    class Meta:
        name = 'multiple_text'
        verbose_name = 'Multiple (Text)'
        multiple = True
        widget_class = MultipleTextWidget

    class OptionsForm(forms.Form):
        min = forms.IntegerField(required=False)
        min_length = forms.IntegerField(required=False)
        max_length = forms.IntegerField(required=False)

    def clean_question_options(self, question_options):
        if 'min' in question_options:
            try:
                question_options['min'] = int(question_options['min'])
            except ValueError:
                raise forms.ValidationError('value for "min" is not an integer')
        if 'min_length' in question_options:
            try:
                question_options['min_length'] = int(question_options['min_length'])
            except ValueError:
                raise forms.ValidationError('value for "min_length" is not an integer')
        if 'max_length' in question_options:
            try:
                question_options['max_length'] = int(question_options['max_length'])
            except ValueError:
                raise forms.ValidationError('value for "max_length" is not an integer')

        return question_options

    def formfield_widget_attrs(self):
        attrs = dict()
        if self.question.question_options.get('min_length'):
            attrs.update({'minlength': self.question.question_options.get('min_length')})
        if self.question.question_options.get('max_length'):
            attrs.update({'maxlength': self.question.question_options.get('max_length')})
        attrs.update(super().formfield_widget_attrs())
        return attrs

    def formfield(self, result_set, min_fields=1):
        fields = list()
        field_widgets = list()

        min_required = self.question.question_options.get('min', 1 if self.question.required else 0)

        prev_answers = result_set.answers.filter(question=self.question).count()
        num_fields = max(prev_answers, min_required, min_fields)

        for field_num in range(0, num_fields):
            # validate required fields in MultipleTextField.clean()
            field = forms.CharField(
                required=False,
                min_length=self.question.question_options.get('min_length'),
                max_length=self.question.question_options.get('max_length'),
            )
            fields.append(field)
            # setup widget
            attrs = self.formfield_widget_attrs()
            # workaround to set required via js (required will be ignored by django if require_all_fields is True)
            if field_num == 0 and self.question.required:
                attrs.update({'data-required': True})
            field_widgets.append(widgets.TextInput(attrs=attrs))

        return MultipleTextField(
            fields=fields,
            widget=self.formfield_widget(widgets=field_widgets),
            required=False,
            require_all_fields=self.question.required,
            label=self.question.question_text,
            question_type=self,
        )



from collections import OrderedDict

from django.contrib.postgres.fields import JSONField
from django.db import models

from ordered_model.models import OrderedModel

from ..fields import DynamicStorageFileField
from .questionnaire import Question, Questionnaire


class QuestionnaireResultManager(models.Manager):
    def with_answers(self):
        return self.get_queryset().prefetch_related('answers')


class QuestionnaireResult(models.Model):
    questionnaire = models.ForeignKey(Questionnaire, on_delete=models.CASCADE)
    result_meta = JSONField(
        default=dict,
        help_text='Optional JSON field for application specific meta data',
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = QuestionnaireResultManager()

    def initial_form_data(self):
        initial_data = OrderedDict()

        initial_data['result_set'] = self.pk
        for question in self.questionnaire.questions.all():
            form_field_id = 'q{}'.format(question.pk)
            form_field_value = question.question_type_obj.initial_field_value(self)
            if form_field_value is not None:
                initial_data[form_field_id] = form_field_value

        return initial_data

    def __str__(self):
        return u'{} ({})'.format(self.questionnaire, self.created_at)


class QuestionAnswer(OrderedModel):
    result_set = models.ForeignKey(QuestionnaireResult, related_name='answers', on_delete=models.CASCADE)
    question = models.ForeignKey(Question, related_name='answers', on_delete=models.CASCADE)
    answer_data = JSONField(null=True, db_index=True)

    order_with_respect_to = 'result_set'

    class Meta(OrderedModel.Meta):
        pass

    def __str__(self):
        return '{}: {}'.format(self.question, self.answer_data)


class AnswerFile(OrderedModel):
    answer = models.OneToOneField(QuestionAnswer, related_name='file_upload', on_delete=models.CASCADE)
    file = DynamicStorageFileField()

    order_with_respect_to = 'answer'

    class Meta(OrderedModel.Meta):
        pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-10 10:16
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import questionnaire_core.fields

try:
    # there is no difference in the (postgres) schema, so we can easily swap between the two
    from django.db.models import JSONField
except ImportError:
    from django.contrib.postgres.fields.jsonb import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AnswerFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField(db_index=True, editable=False)),
                ('file', questionnaire_core.fields.DynamicStorageFileField()),
            ],
            options={
                'ordering': ('order',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Question',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField(db_index=True, editable=False)),
                ('question_text', models.CharField(max_length=5000)),
                ('question_type', questionnaire_core.fields.QuestionTypeField(max_length=32)),
                ('question_options', JSONField(blank=True, default=dict)),
                ('required', models.BooleanField(default=True)),
            ],
            options={
                'ordering': ('order',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='QuestionAnswer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveIntegerField(db_index=True, editable=False)),
                ('answer_data', JSONField(db_index=True, null=True)),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='answers', to='questionnaire_core.Question')),
            ],
            options={
                'ordering': ('order',),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Questionnaire',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='QuestionnaireResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('result_meta', JSONField(default=dict, help_text='Optional JSON field for application specific meta data')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('questionnaire', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='questionnaire_core.Questionnaire')),
            ],
        ),
        migrations.AddField(
            model_name='questionanswer',
            name='result_set',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='answers', to='questionnaire_core.QuestionnaireResult'),
        ),
        migrations.AddField(
            model_name='question',
            name='questionnaire',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='questions', to='questionnaire_core.Questionnaire'),
        ),
        migrations.AddField(
            model_name='answerfile',
            name='answer',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='file_upload', to='questionnaire_core.QuestionAnswer'),
        ),
    ]



from decimal import Decimal

from django import forms

from .base import QuestionTypeBase


class NumberBase(QuestionTypeBase):
    class Meta:
        abstract = True

    class OptionsForm(forms.Form):
        min = forms.IntegerField(required=False)
        max = forms.IntegerField(required=False)

    def clean_question_options(self, question_options):
        """
        expected question_options format:
        {
            min: 1,
            max: 10
        }
        """

        if 'min' in question_options:
            try:
                question_options['min'] = int(question_options['min'])
            except ValueError:
                raise forms.ValidationError('value for "min" is not an integer')

        if 'max' in question_options:
            try:
                question_options['max'] = int(question_options['max'])
            except ValueError:
                raise forms.ValidationError('value for "max" is not an integer')

        return question_options


class NumberInteger(NumberBase):
    class Meta:
        name = 'number_integer'
        verbose_name = 'Number (Integer)'
        widget_class = forms.NumberInput

    class OptionsForm(NumberBase.OptionsForm):
        pass

    def formfield(self, result_set):
        min_value = self.question.question_options.get('min')
        max_value = self.question.question_options.get('max')

        return forms.IntegerField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            min_value=min_value,
            max_value=max_value,
        )


class NumberDecimal(NumberBase):
    class Meta:
        name = 'number_decimal'
        verbose_name = 'Number (Decimal)'
        widget_class = forms.NumberInput

    class OptionsForm(NumberBase.OptionsForm):
        decimal_places = forms.IntegerField(required=False)

    def clean_question_options(self, question_options):
        """
        expected question_options format:
        {
            min: 1,
            max: 10,
            decimal_places: 2
        }
        """

        question_options = super().clean_question_options(question_options)

        if 'decimal_places' in question_options:
            try:
                question_options['decimal_places'] = int(question_options['decimal_places'])
            except ValueError:
                raise forms.ValidationError('value for "decimal_places" is not an integer')

        return question_options

    def clean_answer_data(self, data):
        if data is not None:
            return str(data.quantize(Decimal('.00')))

    def formfield(self, result_set):
        min_value = self.question.question_options.get('min')
        max_value = self.question.question_options.get('max')
        decimal_places = self.question.question_options.get('decimal_places', 2)

        return forms.DecimalField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            min_value=min_value,
            max_value=max_value,
            decimal_places=decimal_places,
        )


class NumberPercent(NumberBase):
    class Meta:
        name = 'number_percent'
        verbose_name = 'Number (Percent)'
        widget_class = forms.NumberInput

    class OptionsForm(NumberBase.OptionsForm):
        pass

    def formfield(self, result_set):
        min_value = self.question.question_options.get('min', 0)
        max_value = self.question.question_options.get('max', 100)

        return forms.IntegerField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            min_value=min_value,
            max_value=max_value,
        )



from django import forms

from .base import QuestionTypeBase


class ChoicesBase(QuestionTypeBase):
    class Meta:
        abstract = True

    class OptionsForm(forms.Form):
        empty_label = forms.CharField(required=False)
        # choices = EditChoicesField()

    def clean_question_options(self, question_options):
        """
        expected question_options format:
        {
            choices: [
                {value: "", label: ""}, ...
            ]
        }
        """
        if 'choices' not in question_options:
            raise forms.ValidationError('key "choices" required')
        if not isinstance(question_options.get('choices'), list):
            raise forms.ValidationError('list in "choices" required')
        if len(question_options.get('choices')) < 1:
            raise forms.ValidationError('list of "choices" is empty')
        for choice in question_options.get('choices'):
            if 'label' not in choice:
                raise forms.ValidationError('key "label" in choice is required')
            if 'value' not in choice:
                raise forms.ValidationError('key "value" in choice is required')

            # Clean whitespace and set 'value' to 'label' if it is empty.
            value = choice['value'].strip()
            label = choice['label'].strip()
            choice['value'] = value if value else label
            choice['label'] = label

        return question_options


class Choices(ChoicesBase):
    class Meta:
        name = 'choices'
        verbose_name = 'Choices'
        widget_class = forms.Select

    def formfield(self, result_set):
        choices = [(c.get('value'), c.get('label')) for c in self.question.question_options.get('choices')]

        if self.question.question_options.get('empty_label'):
            choices = [('', self.question.question_options.get('empty_label'))] + choices

        return forms.ChoiceField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            choices=choices,
            # validators=[validators.validate_slug],
        )


class ChoicesMultiple(ChoicesBase):
    class Meta:
        name = 'choices_multiple'
        verbose_name = 'Choices multiple'
        multiple = True
        widget_class = forms.CheckboxSelectMultiple

    def formfield(self, result_set):
        choices = [(c.get('value'), c.get('label')) for c in self.question.question_options.get('choices')]

        if self.question.question_options.get('empty_label'):
            choices = [('', self.question.question_options.get('empty_label'))] + choices

        return forms.MultipleChoiceField(
            widget=self.formfield_widget(),
            label=self.question.question_text,
            required=self.question.required,
            choices=choices,
            # validators=[validators.validate_slug],
        )



from .questionnaire import Question, Questionnaire
from .result import AnswerFile, QuestionAnswer, QuestionnaireResult


__all__ = (
    'Questionnaire',
    'Question',
    'QuestionnaireResult',
    'QuestionAnswer',
    'AnswerFile',
)



from django.apps import AppConfig


class QuestionnaireCoreConfig(AppConfig):
    name = 'questionnaire_core'

    def ready(self):
        # validate settings on startup
        from . import settings
        # update question type choices on `Question` model
        from .models import Question
        question_type_choices = [(qt.meta.name, qt.meta.verbose_name) for qt in settings.active_question_types]
        question_type_field = Question._meta.get_field('question_type')
        question_type_field.choices = question_type_choices
